@@ -20,6 +20,8 @@ import os.path
 import threading
 import uuid
 
+from alembic import migration
+from alembic import operations
 import oslo_db.api
 from oslo_db import exception
 from oslo_db.sqlalchemy import enginefacade
@@ -106,8 +108,10 @@ class PerInstanceFacade(object):
 
 class ResourceClassMapper(object):
     def __init__(self):
+        # FIXME(sileht): 3 attributes, perhaps we need a better structure.
         self._cache = {'generic': {'resource': base.Resource,
-                                   'history': base.ResourceHistory}}
+                                   'history': base.ResourceHistory,
+                                   'updated_at': utils.utcnow()}}
 
     @staticmethod
     def _build_class_mappers(resource_type, baseclass=None):
@@ -127,14 +131,24 @@ class ResourceClassMapper(object):
             {"__tablename__": ("%s_history" % tablename),
              "__table_args__": tables_args})
         return {'resource': resource_ext,
-                'history': resource_history_ext}
+                'history': resource_history_ext,
+                'updated_at': resource_type.updated_at}
 
     def get_classes(self, resource_type):
         # NOTE(sileht): We don't care about concurrency here because we allow
         # sqlalchemy to override its global object with extend_existing=True
         # this is safe because classname and tablename are uuid.
         try:
-            return self._cache[resource_type.tablename]
+            mappers = self._cache[resource_type.tablename]
+            # Cache is outdated
+            if (resource_type.name != "generic"
+                    and resource_type.updated_at > mappers['updated_at']):
+                for table_purpose in ['resource', 'history']:
+                    Base.metadata.remove(Base.metadata.tables[
+                        mappers[table_purpose].__tablename__])
+                del self._cache[resource_type.tablename]
+                raise KeyError
+            return mappers
         except KeyError:
             mapper = self._build_class_mappers(resource_type)
             self._cache[resource_type.tablename] = mapper
@@ -147,8 +161,8 @@ class ResourceClassMapper(object):
                                "creating")
 
         mappers = self.get_classes(resource_type)
-        tables = [Base.metadata.tables[klass.__tablename__]
-                  for klass in mappers.values()]
+        tables = [Base.metadata.tables[mappers["resource"].__tablename__],
+                  Base.metadata.tables[mappers["history"].__tablename__]]
 
         try:
             with facade.writer_connection() as connection:
@@ -179,7 +193,7 @@ class ResourceClassMapper(object):
                 and inn_e.orig.pgcode == '25P02')
 
     @retry_on_deadlock
-    def unmap_and_delete_tables(self, resource_type, connection):
+    def unmap_and_delete_tables(self, resource_type, facade):
         if resource_type.state != "deleting":
             raise RuntimeError("unmap_and_delete_tables must be called in "
                                "state deleting")
@@ -187,18 +201,18 @@ class ResourceClassMapper(object):
         mappers = self.get_classes(resource_type)
         del self._cache[resource_type.tablename]
 
-        tables = [Base.metadata.tables[klass.__tablename__]
-                  for klass in mappers.values()]
+        tables = [Base.metadata.tables[mappers['resource'].__tablename__],
+                  Base.metadata.tables[mappers['history'].__tablename__]]
 
-        if connection is not None:
-            # NOTE(sileht): Base.metadata.drop_all doesn't
-            # issue CASCADE stuffs correctly at least on postgresql
-            # We drop foreign keys manually to not lock the destination
-            # table for too long during drop table.
-            # It's safe to not use a transaction since
-            # the resource_type table is already cleaned and commited
-            # so this code cannot be triggerred anymore for this
-            # resource_type
+        # NOTE(sileht): Base.metadata.drop_all doesn't
+        # issue CASCADE stuffs correctly at least on postgresql
+        # We drop foreign keys manually to not lock the destination
+        # table for too long during drop table.
+        # It's safe to not use a transaction since
+        # the resource_type table is already cleaned and committed
+        # so this code cannot be triggerred anymore for this
+        # resource_type
+        with facade.writer_connection() as connection:
             try:
                 for table in tables:
                     for fk in table.foreign_key_constraints:
@@ -221,13 +235,9 @@ class ResourceClassMapper(object):
 
             # NOTE(sileht): If something goes wrong here, we are currently
             # fucked, that why we expose the state to the superuser.
-            # TODO(sileht): The idea is to make the delete resource_type more
-            # like a cleanup method, I mean we should don't fail if the
-            # constraint have already been dropped or the table have already
-            # been deleted. So, when the superuser have fixed it's backend
-            # issue, it can rerun 'DELETE ../resource_type/foobar' even the
-            # state is already error and if we are sure all underlying
-            # resources have been cleaned we really deleted the resource_type.
+            # But we allow him to delete a resource type in error state
+            # in case of he cleanup the mess manually and want gnocchi to
+            # control and finish the cleanup.
 
         # TODO(sileht): Remove this resource on other workers
         # by using expiration on cache ?
@@ -376,6 +386,58 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         resource_type.state = "active"
         return resource_type
 
+    def update_resource_type(self, name, add_attributes=None,
+                             del_attributes=None):
+        if not add_attributes and not del_attributes:
+            return
+        add_attributes = add_attributes or []
+        del_attributes = del_attributes or []
+
+        self._set_resource_type_state(name, "updating", "active")
+
+        try:
+            with self.facade.independent_writer() as session:
+                rt = self._get_resource_type(session, name)
+
+                with self.facade.writer_connection() as connection:
+                    ctx = migration.MigrationContext.configure(connection)
+                    op = operations.Operations(ctx)
+                    for table in [rt.tablename, '%s_history' % rt.tablename]:
+                        with op.batch_alter_table(table) as batch_op:
+                            for attr in del_attributes:
+                                batch_op.drop_column(attr)
+                            for attr in add_attributes:
+                                # TODO(sileht): When attr.required is True, we
+                                # have to pass a default. rest layer current
+                                # protect us, requied = True is not yet allowed
+                                batch_op.add_column(sqlalchemy.Column(
+                                    attr.name, attr.satype,
+                                    nullable=not attr.required))
+
+                rt.state = "active"
+                rt.updated_at = utils.utcnow()
+                rt.attributes.extend(add_attributes)
+                for attr in list(rt.attributes):
+                    if attr.name in del_attributes:
+                        rt.attributes.remove(attr)
+                # FIXME(sileht): yeah that's wierd but attributes is a custom
+                # json column and 'extend' doesn't trigger sql update, this
+                # enforce the update. I wonder if sqlalchemy provides something
+                # on column description side.
+                sqlalchemy.orm.attributes.flag_modified(rt, 'attributes')
+
+        except Exception:
+            # NOTE(sileht): We fail the DDL, we have no way to automatically
+            # recover, just set a particular state
+            # TODO(sileht): Create a repair REST endpoint that delete
+            # columns not existing in the database but in the resource type
+            # description. This will allow to pass wrong update_error to active
+            # state, that currently not possible.
+            self._set_resource_type_state(name, "updating_error")
+            raise
+
+        return rt
+
     def get_resource_type(self, name):
         with self.facade.independent_reader() as session:
             return self._get_resource_type(session, name)
@@ -387,12 +449,20 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         return resource_type
 
     @retry_on_deadlock
-    def _set_resource_type_state(self, name, state):
+    def _set_resource_type_state(self, name, state,
+                                 expected_previous_state=None):
         with self.facade.writer() as session:
             q = session.query(ResourceType)
             q = q.filter(ResourceType.name == name)
+            if expected_previous_state is not None:
+                q = q.filter(ResourceType.state == expected_previous_state)
             update = q.update({'state': state})
             if update == 0:
+                if expected_previous_state is not None:
+                    rt = session.query(ResourceType).get(name)
+                    if rt:
+                        raise indexer.UnexpectedResourceTypeState(
+                            name, expected_previous_state, rt.state)
                 raise indexer.IndexerException(
                     "Fail to set resource type state of %s to %s" %
                     (name, state))
@@ -421,9 +491,12 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         try:
             with self.facade.writer() as session:
                 rt = self._get_resource_type(session, name)
-                if rt.state != "active":
+                if rt.state not in ["active", "deletion_error",
+                                    "creation_error", "updating_error"]:
                     raise indexer.UnexpectedResourceTypeState(
-                        name, "active", rt.state)
+                        name,
+                        "active/deletion_error/creation_error/updating_error",
+                        rt.state)
                 session.delete(rt)
 
                 # FIXME(sileht): Why do I need to flush here !!!
@@ -464,9 +537,8 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         rt = self._mark_as_deleting_resource_type(name)
 
         try:
-            with self.facade.writer_connection() as connection:
-                self._RESOURCE_TYPE_MANAGER.unmap_and_delete_tables(
-                    rt, connection)
+            self._RESOURCE_TYPE_MANAGER.unmap_and_delete_tables(
+                rt, self.facade)
         except Exception:
             # NOTE(sileht): We fail the DDL, we have no way to automatically
             # recover, just set a particular state
@@ -475,7 +547,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
         self._delete_resource_type(name)
 
-    def _resource_type_to_classes(self, session, name):
+    def _resource_type_to_mappers(self, session, name):
         resource_type = self._get_resource_type(session, name)
         if resource_type.state != "active":
             raise indexer.UnexpectedResourceTypeState(
@@ -643,7 +715,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             raise ValueError(
                 "Start timestamp cannot be after end timestamp")
         with self.facade.writer() as session:
-            resource_cls = self._resource_type_to_classes(
+            resource_cls = self._resource_type_to_mappers(
                 session, resource_type)['resource']
             r = resource_cls(
                 id=id,
@@ -679,9 +751,9 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                         create_revision=True,
                         **kwargs):
         with self.facade.writer() as session:
-            classes = self._resource_type_to_classes(session, resource_type)
-            resource_cls = classes["resource"]
-            resource_history_cls = classes["history"]
+            mappers = self._resource_type_to_mappers(session, resource_type)
+            resource_cls = mappers["resource"]
+            resource_history_cls = mappers["history"]
 
             try:
                 # NOTE(sileht): We use FOR UPDATE that is not galera friendly,
@@ -797,10 +869,41 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                     Resource.id == resource_id).delete() == 0:
                 raise indexer.NoSuchResource(resource_id)
 
+    @retry_on_deadlock
+    def delete_resources(self, resource_type='generic',
+                         attribute_filter=None):
+        if not attribute_filter:
+            raise ValueError("attribute_filter must be set")
+
+        with self.facade.writer() as session:
+            target_cls = self._resource_type_to_mappers(
+                session, resource_type)["resource"]
+
+            q = session.query(target_cls.id)
+
+            engine = session.connection()
+            try:
+                f = QueryTransformer.build_filter(engine.dialect.name,
+                                                  target_cls,
+                                                  attribute_filter)
+            except indexer.QueryAttributeError as e:
+                # NOTE(jd) The QueryAttributeError does not know about
+                # resource_type, so convert it
+                raise indexer.ResourceAttributeError(resource_type,
+                                                     e.attribute)
+
+            q = q.filter(f)
+
+            session.query(Metric).filter(
+                Metric.resource_id.in_(q)
+            ).update({"status": "delete"},
+                     synchronize_session=False)
+            return q.delete(synchronize_session=False)
+
     @retry_on_deadlock
     def get_resource(self, resource_type, resource_id, with_metrics=False):
         with self.facade.independent_reader() as session:
-            resource_cls = self._resource_type_to_classes(
+            resource_cls = self._resource_type_to_mappers(
                 session, resource_type)['resource']
             q = session.query(
                 resource_cls).filter(
@@ -810,9 +913,9 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             return q.first()
 
     def _get_history_result_mapper(self, session, resource_type):
-        classes = self._resource_type_to_classes(session, resource_type)
-        resource_cls = classes['resource']
-        history_cls = classes['history']
+        mappers = self._resource_type_to_mappers(session, resource_type)
+        resource_cls = mappers['resource']
+        history_cls = mappers['history']
 
         resource_cols = {}
         history_cols = {}
@@ -864,7 +967,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                 target_cls = self._get_history_result_mapper(
                     session, resource_type)
             else:
-                target_cls = self._resource_type_to_classes(
+                target_cls = self._resource_type_to_mappers(
                     session, resource_type)["resource"]
 
             q = session.query(target_cls)
@@ -917,7 +1020,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                         all_resources.extend(resources)
                     else:
                         try:
-                            target_cls = self._resource_type_to_classes(
+                            target_cls = self._resource_type_to_mappers(
                                 session, type)['history' if is_history else
                                                'resource']
                         except (indexer.UnexpectedResourceTypeState,
@@ -1080,7 +1183,11 @@ class QueryTransformer(object):
 
                 if converter:
                     try:
-                        value = converter(value)
+                        if isinstance(value, list):
+                            # we got a list for in_ operator
+                            value = [converter(v) for v in value]
+                        else:
+                            value = converter(value)
                     except Exception:
                         raise indexer.QueryValueError(value, field_name)
 
